import os
import abc
import sys
import inspect
import pkgutil
from contextlib import contextmanager, _GeneratorContextManager
from typing import Any, Callable, Optional, List

from . import exceptions
from .util.swig import WrapperMetaClass, pointer_class, ContextManagedPointerClass
from .util.calculate import calculate
from .esys_binding import *
from .fapi_binding import *

# MODULE_NAME = "tpm2_pytss"
MODULE_NAME = ".".join(__name__.split(".")[:-1])
esys_binding = sys.modules[MODULE_NAME + ".esys_binding"]
fapi_binding = sys.modules[MODULE_NAME + ".fapi_binding"]

# TODO signature with return type of TPM_RC
for prefix, module in [("Esys_", esys_binding), ("Fapi_", fapi_binding)]:
    exceptions.wrap_funcs(
        module,
        dst=[module, sys.modules[__name__]],
        cond=lambda key, func: key.startswith(prefix) or key.startswith("Tss2_"),
    )

    exceptions.wrap_funcs(
        module,
        dst=[module, sys.modules[__name__]],
        cond=lambda key, func: key.startswith("Tss2_MU_"),
        wrapperfunc=exceptions.raise_tpm2_mu_error,
    )


class NoPointerClass(Exception):
    """
    Happens when there is no pointer for a SWIG value, shouldn't ever happen.
    """

    pass  # pragma: no cov


def wrap_all_matching(
    wrapper_func: Callable[[str, Any], Callable],
    src,
    dst: Optional[List[Any]] = None,
    cond: Optional[Callable[[Callable], bool]] = None,
):
    """
    Wrap the functions within the given module with a decorator so that if they
    return integer values that are non-zero an error will be thrown.
    """
    if dst is None:
        dst = [src]
    for key, value in inspect.getmembers(src):
        if cond is None or cond(key, value):
            value = wrapper_func(key, value)
            for obj in dst:
                setattr(obj, key, value)
    return dst


class SetPropertiesViaInit:
    """
    SWIG creates python classes but doesn't let us set their properties when
    passing arguments to their ``__init__`` method. This class is meant to
    subclass from those swig classes to allow for that functionality.
    """

    def __init__(self, **kwargs):
        sig = inspect.signature(self.SWIG_CLS)
        if sig.parameters:
            self.SWIG_CLS.__init__(self, **kwargs)
        else:
            self.SWIG_CLS.__init__(self)
        for key, value in kwargs.items():
            # For arrays which are not BYTEs within structures we check if
            # attempting to set the value threw an exception complaining about
            # the wrong data type. If it did and the value is a list or tuple,
            # then set each index.
            try:
                if isinstance(value, dict) and not len(value):
                    continue
                setattr(self, key, value)
            except TypeError as error:
                if not isinstance(value, (tuple, list, set, bytes, bytearray)):
                    error.args = (error.args[0] + " but got %r" % (value),)
                    raise
                str_error = str(error)
                if not "argument 2 of type" in str_error:
                    raise
                str_error = str_error.split("argument 2 of type")
                str_error = str_error[-1].replace("'", "")
                str_error = str_error.strip()
                # Split the type name from the size of the array
                type_name, max_size = str_error.split(" ")
                if not (max_size.startswith("[") and max_size[::-1].startswith("]")):
                    raise
                max_size = calculate(
                    sys.modules[__name__], max_size.replace("[", "").replace("]", "")
                )
                # Grab the array verion of the type
                array_cls = getattr(sys.modules[__name__], type_name + "_ARRAY")
                # Create the array version and assign to each index
                array = array_cls.frompointer(getattr(self, key))
                # Copy each index
                for i in range(0, min(len(value), max_size)):
                    array[i] = value[i]

    @contextmanager
    def ptr(self):
        ptr_cls = ESYSBinding.MODULE.__dict__.get("{}_PTR".format(self.NAME), None)
        if ptr_cls is None:
            raise NoPointerClass(self.NAME)
        # TODO If the value of the pointer changes then we need to set the
        # properites of this object to be whatever they were changed to in the
        # pointers object (Since we can't accually point to this value)
        with ptr_cls(self) as ptr:
            yield ptr


def set_properties_via_init(name, cls):
    """
    SWIG creates python classes but doesn't let us set their properties when
    passing arguments to their ``__init__`` method. Grab the class and use
    it and SetPropertiesViaInit to create a new class which will allow for
    that functionality.
    """
    return type(name, (SetPropertiesViaInit, cls), {"NAME": name, "SWIG_CLS": cls})


for module in [esys_binding, fapi_binding]:
    wrap_all_matching(
        set_properties_via_init,
        module,
        dst=[module, sys.modules[__name__]],
        cond=lambda key, value: inspect.isclass(value) and key.upper() == key,
    )


class ByteArrayHelper:
    def __init__(self, **kwargs):
        # kwargs which need to be copied from a list or bytearry into a C array
        convert_buffer = {}
        for size_prop, buffer_prop, buffer_max in self.BYTEARRAYS:
            if buffer_prop in kwargs:
                # Skip if not a python type
                if not isinstance(kwargs[buffer_prop], (tuple, list, bytes, bytearray)):
                    continue
                if buffer_prop == "pcrSelections":
                    continue
                # Add to buffers to be converted
                convert_buffer[buffer_prop] = kwargs[buffer_prop]
                # Remove from kwargs so they don't get set
                del kwargs[buffer_prop]
            if size_prop in kwargs:
                # Remove from kwargs so they don't get set
                del kwargs[size_prop]
        self.BYTEARRAY_CLS.__init__(self, **kwargs)
        for size_prop, buffer_prop, buffer_max in self.BYTEARRAYS:
            # Skip buffers we weren't given
            if buffer_prop not in convert_buffer:
                continue
            buffer_data = convert_buffer[buffer_prop]
            if len(buffer_data) > buffer_max:
                raise ValueError(
                    "{} of size {} is larger than maximum of {}".format(
                        buffer_prop, len(buffer_data), buffer_max
                    )
                )
            # Set size
            setattr(self, size_prop, len(buffer_data))
            # Set elements of buffer
            buffer_ptr = ByteArray.frompointer(getattr(self, buffer_prop))
            for i, value in enumerate(buffer_data):
                buffer_ptr[i] = value


BYTEARRAY_STRUCTURES = {
    "TPM2B_DIGEST": [("size", "buffer", sizeof_TPMU_HA)],
    "TPM2B_DATA": [("size", "buffer", sizeof_TPMU_HA)],
    "TPM2B_EVENT": [("size", "buffer", 1024)],
    "TPM2B_MAX_BUFFER": [("size", "buffer", TPM2_MAX_DIGEST_BUFFER)],
    "TPM2B_MAX_NV_BUFFER": [("size", "buffer", TPM2_MAX_NV_BUFFER_SIZE)],
    "TPM2B_IV": [("size", "buffer", TPM2_MAX_SYM_BLOCK_SIZE)],
    "TPM2B_NAME": [("size", "name", sizeof_TPMU_NAME)],
    "TPMS_PCR_SELECT": [("sizeofSelect", "pcrSelect", TPM2_PCR_SELECT_MAX)],
    "TPMS_PCR_SELECTION": [("sizeofSelect", "pcrSelect", TPM2_PCR_SELECT_MAX)],
    "TPMS_TAGGED_PCR_SELECT": [("sizeofSelect", "pcrSelect", TPM2_PCR_SELECT_MAX)],
    "TPML_CC": [("count", "commandCodes", TPM2_MAX_CAP_CC)],
    "TPML_CCA": [("count", "commandAttributes", TPM2_MAX_CAP_CC)],
    "TPML_ALG": [("count", "algorithms", TPM2_MAX_ALG_LIST_SIZE)],
    "TPML_HANDLE": [("count", "handle", TPM2_MAX_CAP_HANDLES)],
    "TPML_DIGEST": [("count", "digests", 8)],
    "TPML_DIGEST_VALUES": [("count", "digests", TPM2_NUM_PCR_BANKS)],
    "TPML_PCR_SELECTION": [("count", "pcrSelections", TPM2_NUM_PCR_BANKS)],
    "TPML_ALG_PROPERTY": [("count", "algProperties", TPM2_MAX_CAP_ALGS)],
    "TPML_TAGGED_TPM_PROPERTY": [("count", "tpmProperty", TPM2_MAX_TPM_PROPERTIES)],
    "TPML_TAGGED_PCR_PROPERTY": [("count", "pcrProperty", TPM2_MAX_PCR_PROPERTIES)],
    "TPML_ECC_CURVE": [("count", "eccCurves", TPM2_MAX_ECC_CURVES)],
    "TPML_INTEL_PTT_PROPERTY": [("count", "property", TPM2_MAX_PTT_PROPERTIES)],
    "TPM2B_ATTEST": [("size", "attestationData", sizeof_TPMS_ATTEST)],
    "TPM2B_SYM_KEY": [("size", "buffer", TPM2_MAX_SYM_KEY_BYTES)],
    "TPM2B_SENSITIVE_DATA": [("size", "buffer", TPM2_MAX_SYM_DATA)],
    "TPM2B_PUBLIC_KEY_RSA": [("size", "buffer", TPM2_MAX_RSA_KEY_BYTES)],
    "TPM2B_PRIVATE_KEY_RSA": [("size", "buffer", TPM2_MAX_RSA_KEY_BYTES / 2)],
    "TPM2B_ECC_PARAMETER": [("size", "buffer", TPM2_MAX_ECC_KEY_BYTES)],
    "TPM2B_ENCRYPTED_SECRET": [("size", "secret", sizeof_TPMU_ENCRYPTED_SECRET)],
    "TPM2B_TEMPLATE": [("size", "buffer", sizeof_TPMT_PUBLIC)],
    "TPM2B_PRIVATE_VENDOR_SPECIFIC": [
        ("size", "buffer", TPM2_PRIVATE_VENDOR_SPECIFIC_BYTES)
    ],
    "TPM2B_PRIVATE": [("size", "buffer", sizeof__PRIVATE)],
    "TPM2B_ID_OBJECT": [("size", "credential", sizeof_TPMS_ID_OBJECT)],
    "TPM2B_CONTEXT_SENSITIVE": [("size", "buffer", TPM2_MAX_CONTEXT_SIZE)],
    "TPM2B_CONTEXT_DATA": [("size", "buffer", sizeof_TPMS_CONTEXT_DATA)],
    "TPML_AC_CAPABILITIES": [("count", "acCapabilities", TPM2_MAX_AC_CAPABILITIES)],
}


def bytearray_helper(name, cls):
    """
    For structures which have internal buffers named "buffer", this function
    modifes those classes so that we can set the buffer from an array and it
    gets returned as a bytearray.
    """
    return type(
        name,
        (ByteArrayHelper, cls),
        {"BYTEARRAY_CLS": cls, "BYTEARRAYS": BYTEARRAY_STRUCTURES[name]},
    )


for module in [esys_binding, fapi_binding]:
    wrap_all_matching(
        bytearray_helper,
        module,
        dst=[module, sys.modules[__name__]],
        cond=lambda key, value: key in BYTEARRAY_STRUCTURES,
    )


def to_bytearray(length, buf):
    """
    Takes any pointer to an array of bytes and that array's length and
    returns a :py:func:`bytearray` containing those bytes.
    """
    buf = ByteArray.frompointer(buf)
    array = bytearray(length)
    for i in range(0, length):
        array[i] = buf[i]
    return array


def var_name(search, str_value):
    """
    Try to find the variable name for the variable ``search`` within the caller
    stack.
    """
    for stack in inspect.stack():
        # Don't care about variable names within tpm2_pytss or contextlib
        # (ExitStack)
        if stack.frame.f_code.co_filename.startswith(
            os.path.dirname(sys.modules[MODULE_NAME].__file__)
        ) or stack.frame.f_code.co_filename.startswith(
            os.path.dirname(sys.modules["contextlib"].__file__)
        ):
            continue
        for variable_name, variable in stack.frame.f_locals.items():
            if variable is search:
                return variable_name
    return str_value


class TRContext(abc.ABC):
    def __init__(self, esys_ctx: "ESYSContext") -> None:
        self.esys_ctx = esys_ctx
        self.ptr = None

    def __enter__(self) -> "TRContext":
        self.ptr = ESYS_TR_PTR()
        self.value = self.esys_ctx.ESYS_TR_NONE
        return self

    @abc.abstractmethod
    def __exit__(self, _exc_type, _exc_value, _traceback):
        pass  # pragma: no cov

    @property
    def value(self):
        return self.ptr.value()

    @value.setter
    def value(self, updated):
        return self.ptr.assign(updated)


@WrapperMetaClass.register_call_mod
def call_mod_tr_context(name, annotation, value):
    # Bail out if the value we're trying to pass isn't a TRContext
    if not isinstance(value, TRContext):
        return
    return WrapperMetaClass.call_mod_ptr_or_value(annotation, value)


class FlushTRContextFlushFailure(Exception):
    """
    Raised when FlushContext results in a TPM2Error. On exit from the
    FlushTRContext.
    """


class FlushTRContext(TRContext):
    def __exit__(self, _exc_type, _exc_value, _traceback):
        if self.value is not None and self.value != self.esys_ctx.ESYS_TR_NONE:
            try:
                self.esys_ctx.FlushContext(self.value)
            except exceptions.TPM2Error as error:
                raise FlushTRContextFlushFailure(var_name(self, self.value)) from error
        self.ptr = False


class AuthSessionContext(FlushTRContext):
    def __init__(self, esys_ctx: "ESYSContext", *args) -> None:
        super().__init__(esys_ctx)
        sig = inspect.signature(self.esys_ctx.StartAuthSession)
        if len(args) != (len(sig.parameters) - 1):
            raise ValueError("Wrong number of arguments")
        for arg, parameter in zip(args, sig.parameters.values()):
            setattr(self, parameter.name, arg)

    def __enter__(self) -> "TRContext":
        super().__enter__()
        self.esys_ctx.StartAuthSession(
            self.tpmKey,
            self.bind,
            self.shandle1,
            self.shandle2,
            self.shandle3,
            self.nonceCaller,
            self.sessionType,
            self.symmetric,
            self.authHash,
            self.ptr,
        )
        return self


class NVContextUndefineSpaceFailure(Exception):
    """
    Raised when NV_UndefineSpace results in a TPM2Error. On exit from the
    NVContext.
    """


# TODO This and AuthSessionContext should be generated on the file using inspect to
# get the arguments of the Define/Undefine and StartAuthSession/Flush
# arguments.
class NVContext(TRContext):
    """
    NV_DefineSpace on enter and NV_UndefineSpace on exit.
    """

    def __init__(
        self,
        esys_ctx: "ESYSContext",
        authHandle=None,
        shandle1=ESYS_TR_NONE,
        shandle2=ESYS_TR_NONE,
        shandle3=ESYS_TR_NONE,
        auth=None,
        publicInfo=None,
    ) -> None:
        super().__init__(esys_ctx)
        self.authHandle = authHandle
        self.shandle1 = shandle1
        self.shandle2 = shandle2
        self.shandle3 = shandle3
        self.auth = auth
        self.publicInfo = publicInfo

    def __enter__(self) -> "TRContext":
        super().__enter__()
        self.esys_ctx.NV_DefineSpace(
            self.authHandle,
            self.shandle1,
            self.shandle2,
            self.shandle3,
            self.auth,
            self.publicInfo,
            self.ptr,
        )
        return self

    def __exit__(self, _exc_type, _exc_value, _traceback):
        if self.value is not None and self.value != self.esys_ctx.ESYS_TR_NONE:
            try:
                self.esys_ctx.NV_UndefineSpace(
                    self.authHandle,
                    self.value,
                    self.shandle1,
                    self.shandle2,
                    self.shandle3,
                )
            except exceptions.TPM2Error as error:
                raise NVContextUndefineSpaceFailure(
                    var_name(self, self.value)
                ) from error
        self.ptr = False


class UsedWithoutEnteringContext(Exception):
    """
    Attempted to use a pointer without being in a ``with`` block for that
    pointer.
    """

    pass  # pragma: no cov


@WrapperMetaClass.register_call_mod
def call_mod_context_managed_pointer_class(name, annotation, value):
    if isinstance(value, _GeneratorContextManager):
        raise UsedWithoutEnteringContext(name)
    if isinstance(value, ContextManagedPointerClass):
        if value.ptr is None:
            raise UsedWithoutEnteringContext(name)
        return WrapperMetaClass.call_mod_ptr_or_value(annotation, value)


# Create all the ContextManagedPointerClasses to be used for the
# pointer_functions
for module in [esys_binding, fapi_binding]:
    for name, func in inspect.getmembers(module):
        if (
            name.startswith("new_")
            and module.__dict__.get(name.replace("new_", "delete_")) is not None
        ):
            name = name[len("new_") :]
            ptr_ptr = pointer_class(name, module=module)
            if ptr_ptr is AttributeError:
                continue
            setattr(sys.modules[__name__], name, ptr_ptr)
            setattr(module, name, ptr_ptr)


def typedef_map():
    """
    SWIG uses typedefs to allow passing a pointer of one type to a function that
    accepts a different type. However, it does not generate aliases for all
    typedefs.

    aka

    typedef A B;

    user would need to create an instance by the name of A when they want to use
    B. This maps B to A so the user can instantiate B without having to know
    that B is the same thing as A.
    """
    return dict(
        [
            line.replace(";", "").split()[-2:][::-1]
            for line in pkgutil.get_data(__name__.split(".")[0], "swig/tpm2_types.i")
            .decode()
            .split("\n")
            if line.startswith("typedef")
        ]
    )


TYPEDEFS = typedef_map()

for module in [esys_binding, fapi_binding]:
    for alias, typename in TYPEDEFS.items():
        for alias, typename in zip(
            [alias, "{}_PTR".format(alias), "{}_PTR_PTR".format(alias)],
            [typename, "{}_PTR".format(typename), "{}_PTR_PTR".format(typename)],
        ):
            # Skip if it already exists
            if sys.modules[__name__].__dict__.get(alias) is not None:
                continue
            reference = sys.modules[__name__].__dict__.get(typename)
            if reference is None:
                continue
            setattr(module, alias, type(alias, (reference,), {}))
            setattr(sys.modules[__name__], alias, type(alias, (reference,), {}))


class ESYSBinding(metaclass=WrapperMetaClass, module=esys_binding):

    MODULE = esys_binding


class FAPIBinding(metaclass=WrapperMetaClass, module=fapi_binding):

    MODULE = fapi_binding



from typing import Optional

from .binding import ESYSBinding
from .util.retry import retry_tcti_loop, retry_tcti_catch


class TCTIContext:
    """
    >>> v = esys.TSS2_ABI_VERSION()
    >>> v.tssCreator = 1
    >>> v.tssFamily = 2
    >>> v.tssLevel = 1
    >>> v.tssVersion = 108
    >>> tcti = esys.TCTI.load("mssim")
    >>> with self.tcti(config=self.tcti_config) as tctx:
    ...     ctxp = esys.new_ctx_ptr()
    ...     esys.Esys_Initialize(ctxp, tctx.ctxp, v)
    """

    def __init__(self, parent, config: str, retry: int = 1):
        self.parent = parent
        self.config = config
        self.retry = retry
        # Context pointer and context pointer pointer
        self.ctxp = None
        self.ctxpp = None

    def __enter__(self):
        ctxpp = ESYSBinding.tcti_ctx_ptr_ptr()
        ctxpp.__enter__()
        # Attempt TCTI connection
        for retry in retry_tcti_loop(max_tries=self.retry):
            with retry_tcti_catch(retry):
                ESYSBinding.Tss2_TctiLdr_Initialize_Ex(
                    self.NAME, self.config, ctxpp.ptr
                )
        # Don't set property until after init succeeds to avoid memory leaks (gc
        # should cleanup if no other references exists when exception is thrown)
        self.ctxp = ctxpp.value
        self.ctxpp = ctxpp
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.ctxpp.__exit__(exc_type, exc_value, traceback)
        self.ctxpp = None
        self.ctxp = None


class TCTI:

    CONTEXT = TCTIContext

    def __call__(
        self, *, config: Optional[str] = None, retry: Optional[int] = 1
    ) -> "TCTIContext":
        return self.CONTEXT(self, config, retry)

    @classmethod
    def _load(cls, name):
        """
        Attemps to load the .so for a tcti given its name. Returns a subclass of
        TCTI.
        """
        return type(
            "{}TCTI".format(name.upper()),
            (cls,),
            {
                "NAME": name,
                "CONTEXT": type(
                    "{}TCTIContext".format(name.upper()), (cls.CONTEXT,), {"NAME": name}
                ),
            },
        )

    @classmethod
    def load(cls, name):
        """
        Attemps to load the .so for a tcti given its name. Returns the
        instantiation of a subclass of TCTI.
        """
        return cls._load(name)()



from contextlib import ExitStack

from tpm2_pytss.binding import *
from tpm2_pytss.util.testing import BaseTestESYS


class TestGetTime(BaseTestESYS):
    def test_get_time(self):
        with ExitStack() as stack:

            authValuePrimary = TPM2B_AUTH(size=5, buffer=[1, 2, 3, 4, 5])

            inSensitivePrimary = TPM2B_SENSITIVE_CREATE(
                size=0,
                sensitive=TPMS_SENSITIVE_CREATE(
                    userAuth=TPM2B_AUTH(size=0, buffer=[0]),
                    data=TPM2B_SENSITIVE_DATA(size=0, buffer=[0]),
                ),
            )

            inSensitivePrimary.sensitive.userAuth = authValuePrimary

            inPublic = TPM2B_PUBLIC(
                size=0,
                publicArea=TPMT_PUBLIC(
                    type=TPM2_ALG_RSA,
                    nameAlg=TPM2_ALG_SHA1,
                    objectAttributes=(
                        TPMA_OBJECT_USERWITHAUTH
                        | TPMA_OBJECT_RESTRICTED
                        | TPMA_OBJECT_SIGN_ENCRYPT
                        | TPMA_OBJECT_FIXEDTPM
                        | TPMA_OBJECT_FIXEDPARENT
                        | TPMA_OBJECT_SENSITIVEDATAORIGIN
                    ),
                    authPolicy=TPM2B_DIGEST(size=0),
                    parameters=TPMU_PUBLIC_PARMS(
                        rsaDetail=TPMS_RSA_PARMS(
                            symmetric=TPMT_SYM_DEF_OBJECT(
                                algorithm=TPM2_ALG_NULL,
                                keyBits=TPMU_SYM_KEY_BITS(aes=128),
                                mode=TPMU_SYM_MODE(aes=TPM2_ALG_CFB),
                            ),
                            scheme=TPMT_RSA_SCHEME(
                                scheme=TPM2_ALG_RSASSA,
                                details=TPMU_ASYM_SCHEME(
                                    rsassa=TPMS_SIG_SCHEME_RSASSA(hashAlg=TPM2_ALG_SHA1)
                                ),
                            ),
                            keyBits=2048,
                            exponent=0,
                        )
                    ),
                    unique=TPMU_PUBLIC_ID(rsa=TPM2B_PUBLIC_KEY_RSA(size=0, buffer={})),
                ),
            )

            authValue = TPM2B_AUTH(size=0, buffer=[])

            outsideInfo = TPM2B_DATA(size=0, buffer=[])

            creationPCR = TPML_PCR_SELECTION(count=0)

            authValue_ptr = stack.enter_context(authValue.ptr())

            self.esys_ctx.TR_SetAuth(ESYS_TR_RH_OWNER, authValue_ptr)

            signHandle = stack.enter_context(self.esys_ctx.flush_tr())

            inSensitivePrimary_ptr = stack.enter_context(inSensitivePrimary.ptr())
            inPublic_ptr = stack.enter_context(inPublic.ptr())
            outsideInfo_ptr = stack.enter_context(outsideInfo.ptr())
            creationPCR_ptr = stack.enter_context(creationPCR.ptr())
            primaryHandle_node_ptr_ptr = stack.enter_context(ESYS_TR_PTR_PTR())
            outPublic_ptr_ptr = stack.enter_context(TPM2B_PUBLIC_PTR_PTR())
            creationData_ptr_ptr = stack.enter_context(TPM2B_CREATION_DATA_PTR_PTR())
            creationHash_ptr_ptr = stack.enter_context(TPM2B_DIGEST_PTR_PTR())
            creationTicket_ptr_ptr = stack.enter_context(TPMT_TK_CREATION_PTR_PTR())

            self.esys_ctx.CreatePrimary(
                ESYS_TR_RH_OWNER,
                ESYS_TR_PASSWORD,
                ESYS_TR_NONE,
                ESYS_TR_NONE,
                inSensitivePrimary_ptr,
                inPublic_ptr,
                outsideInfo_ptr,
                creationPCR_ptr,
                signHandle,
                outPublic_ptr_ptr,
                creationData_ptr_ptr,
                creationHash_ptr_ptr,
                creationTicket_ptr_ptr,
            )

            authValuePrimary_ptr = stack.enter_context(authValuePrimary.ptr())

            self.esys_ctx.TR_SetAuth(signHandle, authValuePrimary_ptr)

            privacyAdminHandle = ESYS_TR_RH_ENDORSEMENT
            inScheme = TPMT_SIG_SCHEME(scheme=TPM2_ALG_NULL)
            qualifyingData = TPM2B_DATA(size=0, buffer=[0])

            qualifyingData_ptr = stack.enter_context(qualifyingData.ptr())
            inScheme_ptr = stack.enter_context(inScheme.ptr())
            timeInfo_ptr_ptr = stack.enter_context(TPM2B_ATTEST_PTR_PTR())
            signature_ptr_ptr = stack.enter_context(TPMT_SIGNATURE_PTR_PTR())

            r = self.esys_ctx.GetTime(
                privacyAdminHandle,
                signHandle,
                ESYS_TR_PASSWORD,
                ESYS_TR_PASSWORD,
                ESYS_TR_NONE,
                qualifyingData_ptr,
                inScheme_ptr,
                timeInfo_ptr_ptr,
                signature_ptr_ptr,
            )
            if (
                (r == TPM2_RC_COMMAND_CODE)
                or (r == (TPM2_RC_COMMAND_CODE | TSS2_RESMGR_RC_LAYER))
                or (r == (TPM2_RC_COMMAND_CODE | TSS2_RESMGR_TPM_RC_LAYER))
            ):
                self.skipTest("Command TPM2_GetTime not supported by TPM")



from contextlib import ExitStack

from tpm2_pytss.binding import *
from tpm2_pytss.util.testing import BaseTestESYS


class TestAutoSessionFlags(BaseTestESYS):
    def test_auto_session_flags(self):
        with ExitStack() as stack:

            symmetric = TPMT_SYM_DEF(
                algorithm=TPM2_ALG_AES,
                keyBits=TPMU_SYM_KEY_BITS(aes=128),
                mode=TPMU_SYM_MODE(aes=TPM2_ALG_CFB),
            )

            nonceCaller = TPM2B_NONCE(
                size=20,
                buffer=[
                    1,
                    2,
                    3,
                    4,
                    5,
                    6,
                    7,
                    8,
                    9,
                    10,
                    11,
                    12,
                    13,
                    14,
                    15,
                    16,
                    17,
                    18,
                    19,
                    20,
                ],
            )

            symmetric_ptr = stack.enter_context(symmetric.ptr())
            nonceCaller_ptr = stack.enter_context(nonceCaller.ptr())

            # Auth session
            session_auth = stack.enter_context(
                self.esys_ctx.auth_session(
                    ESYS_TR_NONE,
                    ESYS_TR_NONE,
                    ESYS_TR_NONE,
                    ESYS_TR_NONE,
                    ESYS_TR_NONE,
                    nonceCaller_ptr,
                    TPM2_SE_HMAC,
                    symmetric_ptr,
                    TPM2_ALG_SHA1,
                )
            )

            # Enc param session
            session_enc = stack.enter_context(
                self.esys_ctx.auth_session(
                    ESYS_TR_NONE,
                    ESYS_TR_NONE,
                    ESYS_TR_NONE,
                    ESYS_TR_NONE,
                    ESYS_TR_NONE,
                    nonceCaller_ptr,
                    TPM2_SE_HMAC,
                    symmetric_ptr,
                    TPM2_ALG_SHA1,
                )
            )

            # Set both ENC and DEC flags for the enc session
            sessionAttributes = (
                TPMA_SESSION_DECRYPT
                | TPMA_SESSION_ENCRYPT
                | TPMA_SESSION_CONTINUESESSION
            )

            self.esys_ctx.TRSess_SetAttributes(session_enc, sessionAttributes, 0xFF)

            auth = TPM2B_AUTH(
                size=20,
                buffer=[
                    10,
                    11,
                    12,
                    13,
                    14,
                    15,
                    16,
                    17,
                    18,
                    19,
                    20,
                    21,
                    22,
                    23,
                    24,
                    25,
                    26,
                    27,
                    28,
                    29,
                ],
            )

            # TODO Move this into binding
            import ctypes

            attributes = ctypes.c_uint32(
                TPMA_NV_OWNERWRITE
                | TPMA_NV_AUTHWRITE
                | TPMA_NV_WRITE_STCLEAR
                | TPMA_NV_READ_STCLEAR
                | TPMA_NV_AUTHREAD
                | TPMA_NV_OWNERREAD
            ).value

            publicInfo = TPM2B_NV_PUBLIC(
                size=0,
                nvPublic=TPMS_NV_PUBLIC(
                    nvIndex=TPM2_NV_INDEX_FIRST,
                    nameAlg=TPM2_ALG_SHA1,
                    attributes=attributes,
                    authPolicy=TPM2B_DIGEST(size=0, buffer=[]),
                    dataSize=20,
                ),
            )

            auth_ptr = stack.enter_context(auth.ptr())
            publicInfo_ptr = stack.enter_context(publicInfo.ptr())

            nvHandle = stack.enter_context(
                self.esys_ctx.nv(
                    authHandle=ESYS_TR_RH_OWNER,
                    shandle1=session_auth,
                    auth=auth_ptr,
                    publicInfo=publicInfo_ptr,
                )
            )

            nv_test_data = TPM2B_MAX_NV_BUFFER(
                size=20,
                buffer=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            )
            nv_test_data_ptr = stack.enter_context(nv_test_data.ptr())

            # NV_Write cmd does not support TPMA_SESSION_ENCRYPT - the flag
            # should be auto cleared by ESYS
            r = self.esys_ctx.NV_Write(
                nvHandle,
                nvHandle,
                session_enc,
                ESYS_TR_NONE,
                ESYS_TR_NONE,
                nv_test_data_ptr,
                0,
            )

            # Verify that the same session flags are still set after the test
            sessionAttributesVerify_ptr = stack.enter_context(TPMA_SESSION_PTR())
            r = self.esys_ctx.TRSess_GetAttributes(
                session_enc, sessionAttributesVerify_ptr
            )

            if sessionAttributes != sessionAttributesVerify_ptr.value:
                raise Exception(
                    "Session flags not equal after write %x, %x"
                    % (sessionAttributes, sessionAttributesVerify_ptr.value)
                )

            data_ptr_ptr = stack.enter_context(TPM2B_MAX_NV_BUFFER_PTR_PTR())

            # NV_Read cmd does not support TPMA_SESSION_DECRYPT - the flags
            # should be auto cleared by ESYS
            r = self.esys_ctx.NV_Read(
                nvHandle,
                nvHandle,
                session_enc,
                ESYS_TR_NONE,
                ESYS_TR_NONE,
                20,
                0,
                data_ptr_ptr,
            )
            # TODO free data_ptr_ptr ?
            # free(data);

            # Verify that the same session flags are still set after the test
            self.esys_ctx.TRSess_GetAttributes(session_enc, sessionAttributesVerify_ptr)

            if sessionAttributes != sessionAttributesVerify_ptr.value:
                raise Exception(
                    "Session flags not equal after read %x, %x"
                    % (sessionAttributes, sessionAttributesVerify_ptr.value)
                )



"""
Helpers for retrying a TCTI connection until it connects successfully. Useful
when the simulator startup is a bit laggy.
"""
import time
import logging
import contextlib

from ..exceptions import TPM2Error

LOGGER = logging.getLogger("tpm2_pytss.util.retry")
TCTI_RETRY_TRIES = 50
TCTI_RETRY_TIMEOUT = 0.5


class TCTIRetry:
    def __init__(
        self, i=0, timeout=TCTI_RETRY_TIMEOUT, tries=0, max_tries=TCTI_RETRY_TRIES
    ):
        self.i = i
        self.timeout = timeout
        self.tries = tries
        self.max_tries = max_tries
        self.success = False

    def __str__(self):
        return "%s(i=%d, timeout=%f, tries=%d, max_tries=%d, success=%s)" % (
            self.__class__.__qualname__,
            self.i,
            self.timeout,
            self.tries,
            self.max_tries,
            self.success,
        )


@contextlib.contextmanager
def retry_tcti_catch(retry):
    retry.success = True
    try:
        yield retry
    except TPM2Error as error:
        retry.success = False
        if not "tcti:IO failure" in str(error):
            raise
        time.sleep(retry.timeout)
        retry.tries += 1
        retry.timeout *= 1.08
        LOGGER.debug(retry)
        if retry.tries > retry.max_tries:
            raise


def retry_tcti_loop(timeout=TCTI_RETRY_TIMEOUT, max_tries=TCTI_RETRY_TRIES):
    retry = TCTIRetry(i=-1, timeout=timeout, max_tries=max_tries)
    while not retry.success:
        retry.i += 1
        yield retry



from contextlib import ExitStack
from tpm2_pytss.binding import *
from tpm2_pytss.exceptions import TPM2Error
from tpm2_pytss.util.testing import BaseTestESYS


class TestTPM2RC(BaseTestESYS):
    def check_error(self, tpm2err, rc, error, parameter, handle, session):
        self.assertEqual(
            tpm2err.rc, rc, "Unexpected RC: %#x, expected %#x" % (tpm2err.rc, rc)
        )
        self.assertEqual(
            tpm2err.error,
            error,
            "Unexpected error: %#x, expected %#x" % (tpm2err.error, error,),
        )
        self.assertEqual(
            tpm2err.parameter,
            parameter,
            "Unexpected parameter: %u, expected %u" % (tpm2err.parameter, parameter,),
        )
        self.assertEqual(
            tpm2err.handle,
            handle,
            "Unexpected handle: %u, expected %u" % (tpm2err.handle, handle,),
        )
        self.assertEqual(
            tpm2err.session,
            session,
            "Unexpected session: %u, expected %u" % (tpm2err.session, session,),
        )

    def test_tpm2_rc_attribute(self):
        with ExitStack() as stack:
            tpm2err = TPM2Error(0)
            try:
                rbytes = b"\x00" * 129
                rnd = TPM2B_SENSITIVE_DATA(buffer=rbytes)
                r = self.esys_ctx.StirRandom(
                    ESYS_TR_NONE, ESYS_TR_NONE, ESYS_TR_NONE, rnd
                )
            except TPM2Error as e:
                tpm2err = e

            self.check_error(
                tpm2err, TPM2_RC_SIZE + TPM2_RC_P + TPM2_RC_1, TPM2_RC_SIZE, 1, 0, 0
            )

    def test_tpm2_rc_handle(self):
        with ExitStack() as stack:
            tpm2err = TPM2Error(0)
            try:
                r = self.esys_ctx.PCR_Reset(
                    ESYS_TR_NONE, ESYS_TR_PASSWORD, ESYS_TR_NONE, ESYS_TR_NONE
                )
            except TPM2Error as e:
                tpm2err = e

            self.check_error(
                tpm2err, TPM2_RC_VALUE + TPM2_RC_H + TPM2_RC_1, TPM2_RC_VALUE, 0, 1, 0
            )

    def test_tpm2_rc_session(self):
        with ExitStack() as stack:
            tpm2err = TPM2Error(0)
            try:
                nbytes = b"\xFF" * 16
                ncaller = TPM2B_NONCE(buffer=nbytes)
                symmetric = TPMT_SYM_DEF(algorithm=TPM2_ALG_NULL)
                shandle = stack.enter_context(self.esys_ctx.flush_tr())
                r = self.esys_ctx.StartAuthSession(
                    ESYS_TR_NONE,
                    ESYS_TR_NONE,
                    ESYS_TR_NONE,
                    ESYS_TR_NONE,
                    ESYS_TR_NONE,
                    ncaller,
                    TPM2_SE_POLICY,
                    symmetric,
                    TPM2_ALG_SHA256,
                    shandle,
                )

                rnd = TPM2B_SENSITIVE_DATA(buffer=nbytes)
                r = self.esys_ctx.StirRandom(shandle, ESYS_TR_NONE, ESYS_TR_NONE, rnd)

            except TPM2Error as e:
                tpm2err = e

            self.check_error(
                tpm2err,
                TPM2_RC_ATTRIBUTES + TPM2_RC_S + TPM2_RC_1,
                TPM2_RC_ATTRIBUTES,
                0,
                0,
                1,
            )



import unittest

from tpm2_pytss.binding import *


class TestBinding(unittest.TestCase):
    def test_set_array(self):
        sel = TPMS_PCR_SELECTION(
            hash=TPM2_ALG_SHA256, sizeofSelect=3, pcrSelect=(0, 0, 255)
        )

        TPMS_PCR_SELECTION(hash=TPM2_ALG_SHA256, sizeofSelect=3, pcrSelect=(0, 0, 255))

        sels = TPML_PCR_SELECTION(count=1, pcrSelections=(sel,))



import random
import tempfile
import contextlib

from tpm2_pytss.fapi import FAPI, FAPIDefaultConfig
from tpm2_pytss.binding import *
from tpm2_pytss.util.simulator import Simulator


def main():
    # Create a context stack
    with contextlib.ExitStack() as ctx_stack:
        # Create a simulator
        simulator = ctx_stack.enter_context(Simulator())
        # Create temporary directories to separate this example's state
        user_dir = ctx_stack.enter_context(tempfile.TemporaryDirectory())
        log_dir = ctx_stack.enter_context(tempfile.TemporaryDirectory())
        system_dir = ctx_stack.enter_context(tempfile.TemporaryDirectory())
        # Create the FAPI object
        fapi = FAPI(
            FAPIDefaultConfig._replace(
                user_dir=user_dir,
                system_dir=system_dir,
                log_dir=log_dir,
                tcti="mssim:port=%d" % (simulator.port,),
                tcti_retry=100,
                ek_cert_less=1,
            )
        )
        # Enter the context, create TCTI connection
        fapi_ctx = ctx_stack.enter_context(fapi)
        # Number of random bytes to get
        length = random.randint(8, 32)
        # Call Fapi_Provision
        fapi_ctx.Provision(None, None, None)
        # Create a pointer to the byte array we'll get back from GetRandom
        array = ctx_stack.enter_context(UINT8_PTR_PTR())
        # Call GetRandom and convert the resulting array to a Python bytearray
        value = to_bytearray(length, fapi_ctx.GetRandom(length, array))
        # Ensure we got the correct number of bytes
        if length != len(value):
            raise AssertionError("Requested %d bytes, got %d" % (length, len(value)))
        # Print bytes to stdout
        print("GetRandom(%d):" % (length,), value)


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: MIT
# Copyright (c) 2019 Intel Corporation
import os
import json
import pathlib
import logging
import tempfile
import contextlib
from functools import partial, wraps
from typing import Optional, ByteString, NamedTuple

from .util.swig import Wrapper
from .util.retry import retry_tcti_loop, retry_tcti_catch
from .binding import (
    AuthSessionContext,
    FAPIBinding,
    FlushTRContext,
    NVContext,
    TPML_PCR_SELECTION_PTR,
    TPMT_HA_PTR,
)
from .context import BaseContextMetaClass
from .config import SYSCONFDIR

# Read default values from system wide default FAPI config
DEFAULT_FAPI_CONFIG_PATH = pathlib.Path(SYSCONFDIR, "tpm2-tss", "fapi-config.json")
DEFAULT_FAPI_CONFIG_CONTENTS = DEFAULT_FAPI_CONFIG_PATH.read_text()
DEFAULT_FAPI_CONFIG = json.loads(DEFAULT_FAPI_CONFIG_CONTENTS)


# Environment variable containing FAPI config
ENV_FAPI_CONFIG = "TSS2_FAPICONF"


FAPIConfig = NamedTuple(
    "FAPIConfig",
    [
        ("profile_dir", str),
        ("user_dir", str),
        ("system_dir", str),
        ("log_dir", str),
        ("profile_name", str),
        ("tcti", str),
        ("system_pcrs", TPML_PCR_SELECTION_PTR),
        ("ek_cert_file", str),
        ("ek_cert_less", bool),
        ("ek_fingerprint", TPMT_HA_PTR),
        ("tcti_retry", int),
    ],
)


def export(self):
    exported = self._asdict()
    remove = [key for key, value in exported.items() if value is None]
    for key in remove:
        del exported[key]
    return exported


@classmethod
def default(cls, **kwargs):
    return cls(
        profile_dir=DEFAULT_FAPI_CONFIG.get("profile_dir", None),
        user_dir=DEFAULT_FAPI_CONFIG.get("user_dir", None),
        system_dir=DEFAULT_FAPI_CONFIG.get("system_dir", None),
        log_dir=DEFAULT_FAPI_CONFIG.get("log_dir", None),
        profile_name=DEFAULT_FAPI_CONFIG.get("profile_name", None),
        tcti=DEFAULT_FAPI_CONFIG.get("tcti", None),
        system_pcrs=DEFAULT_FAPI_CONFIG.get("system_pcrs", None),
        ek_cert_file=DEFAULT_FAPI_CONFIG.get("ek_cert_file", None),
        ek_cert_less=DEFAULT_FAPI_CONFIG.get("ek_cert_less", None),
        ek_fingerprint=DEFAULT_FAPI_CONFIG.get("ek_fingerprint", None),
        tcti_retry=1,
    )


FAPIConfig.export = export
FAPIConfig.default = default
FAPIDefaultConfig = FAPIConfig.default()


class InvalidArgumentError(Exception):
    pass  # pragma: no cov


@contextlib.contextmanager
def temp_fapi_config(config):
    if config is None:
        yield
        return
    with tempfile.NamedTemporaryFile(mode="w") as fileobj:
        old_fapi_config = os.environ.get(ENV_FAPI_CONFIG, None)
        try:
            exported = config.export()
            del exported["tcti_retry"]
            json.dump(exported, fileobj)
            fileobj.seek(0)
            os.environ[ENV_FAPI_CONFIG] = fileobj.name
            yield
        finally:
            if old_fapi_config:
                os.environ[ENV_FAPI_CONFIG] = old_fapi_config


class FAPIMetaClass(BaseContextMetaClass):
    PREFIX = "Fapi_"
    NO_PASS_CTXP = set(["Fapi_Initialize", "Fapi_Finalize", "Fapi_Free"])


class FAPI(Wrapper, metaclass=FAPIMetaClass):

    MODULE = FAPIBinding
    # TODO Verify maximum length requested of GetRandom
    GET_RANDOM_MAX_LENGTH = 64

    def __init__(self, config: Optional[FAPIConfig] = None) -> None:
        self.config = config
        self.ctxpp = None
        self.ctxp = None
        self.logger = logging.getLogger(__name__ + "." + self.__class__.__qualname__)
        self.logger.debug("__init__(%r)", self.config)

    def __enter__(self) -> "FAPIContext":
        # Create an FAPI_CONTEXT **
        ctxpp = self.new_fapi_ctx_ptr()
        # Create an ExitStack
        self.ctx_stack = contextlib.ExitStack().__enter__()
        # Set the config
        self.ctx_stack.enter_context(temp_fapi_config(self.config))
        # Create the FAPI_CONTEXT * and connect to the TPM
        for retry in retry_tcti_loop(max_tries=self.config.tcti_retry):
            with retry_tcti_catch(retry):
                self.Initialize(ctxpp, None)
        # Grab the allocated FAPI_CONTEXT *
        ctxp = self.fapi_ctx_ptr_value(ctxpp)
        # Save references at the end to avoid possible memory leaks
        self.ctxpp = ctxpp
        self.ctxp = ctxp
        # self.logger.debug("__enter__")
        return self

    def __exit__(self, _exc_type, _exc_value, _traceback):
        # Exit the ExitStack
        self.ctx_stack.__exit__(None, None, None)
        # Clean up the FAPI_CONTEXT *
        self.Finalize(self.ctxpp)
        self.delete_fapi_ctx_ptr(self.ctxpp)
        self.ctxpp = None
        self.ctxp = None
        # self.logger.debug("__exit__")

    def _bytearray(self, length, buf):
        """
        Takes any pointer to an array of bytes and that array's length and
        returns a :py:`bytearray` containing those bytes.
        """
        buf = FAPIBinding.ByteArray.frompointer(buf)
        array = bytearray(length)
        for i in range(0, length):
            array[i] = buf[i]
        return array

    @property
    def flush_tr(self):
        """
        Create and return a FAPI_TR_PTR that will be set to FAPI_TR_NONE and
        flushed when it's context exits.
        """
        return partial(FlushTRContext, self)

    @property
    def auth_session(self):
        """
        Create and return a auth session context.
        """
        return partial(AuthSessionContext, self)

    @property
    def nv(self):
        """
        Create and return a nv context.
        """
        return partial(NVContext, self)

    def get_random(self, length: int,) -> ByteString:
        """
        Fapi_GetRandom
        """
        if length > self.GET_RANDOM_MAX_LENGTH:
            raise InvalidArgumentError(
                "Maximum length is {}".format(self.GET_RANDOM_MAX_LENGTH)
            )

        with self.TPM2B_DIGEST_PTR_PTR() as datapp:
            self.GetRandom(length, datapp)

            return self._bytearray(datapp.value.size, datapp.value.buffer)



import os
import inspect
import logging
from functools import partial, wraps
from typing import Any

logging.basicConfig(
    level=getattr(logging, os.environ.get("TPM2_PYTSS_LOG_LEVEL", "CRITICAL").upper())
)
LOGGER = logging.getLogger(__name__)


class PointerAlreadyInUse(Exception):
    pass  # pragma: no cov


class ContextManagedPointerClass:
    """
    By forcing context management we ensure users of the bindings are explicit
    about their usage and freeing of allocated resources. Rather than relying on
    the garbage collector. This makes it harder for them to leave assets lying
    around.
    """

    def __init__(self, value: Any = None):
        self._init_value = value
        self.ptr = None

    @property
    def value(self) -> Any:
        return self._value(self.ptr)

    @value.setter
    def value(self, value) -> None:
        self._assign(self.ptr, value)

    @classmethod
    def frompointer(cls, ptr: Any) -> "ContextManagedPointerClass":
        return cls(ptr)

    def __enter__(self):
        if self.ptr is not None:
            raise PointerAlreadyInUse()
        self.ptr = self._new()
        if self._init_value is not None:
            self.value = self._init_value
        return self

    def __exit__(self, _exc_type, _exc_value, _traceback):
        self._delete(self.ptr)
        self.ptr = None


def pointer_class(name, *, module=None):
    """
    Creates a class of the requested pointer functions data type
    which supports context management.
    """
    check = {
        "_new": "new_{}",
        "_copy": "copy_{}",
        "_delete": "delete_{}",
        "_assign": "{}_assign",
        "_value": "{}_value",
    }
    # Look up the methods
    for key, value in check.items():
        check[key] = module.__dict__.get(value.format(name), None)
    if not all(check.values()):
        return AttributeError
    # Ensure we don't pass self to the functions
    for key, value in check.items():
        check[key] = partial(value)
    return type(name, (ContextManagedPointerClass,), check)


class Wrapper:
    """
    SWIG does a great job. This class takes SWIG outputs and makes them a bit
    more Pythonic.
    """

    def __getattribute__(self, name):
        try:
            return super().__getattribute__(name)
        except AttributeError:
            for attempt in [
                partial(pointer_class, module=self.MODULE),
                lambda name: self.MODULE.__dict__.get(name, AttributeError),
            ]:
                prop = attempt(name)
                if prop is not AttributeError:
                    return prop
            raise


class WrapperMetaClass(type, Wrapper):
    # Enable changing function arguments of one value into another before they
    # are passed to the swig function. This allows us to create abstractions on
    # top of the swig abstractions to make the interface more user friendly.
    CALL_MODS = set()

    def __init__(cls, name, bases, namespace, **kwargs):
        """
        Needed for compatibility with Python 3.5
        """
        super().__init__(name, bases, namespace)

    def __new__(cls, name, bases, props, module=None):
        # Set the module
        props["MODULE"] = module
        # Create the class
        cls = super(WrapperMetaClass, cls).__new__(cls, name, bases, props)
        # Go through all the functions in the module
        for key, func in module.__dict__.items():
            if not key.startswith("_") and inspect.isfunction(func):
                orig = func
                func = cls.wrap(func)
                func = partial(func)
                func.orig = orig
                setattr(cls, key, func)
        return cls

    def __getattribute__(cls, name):
        try:
            return object.__getattribute__(cls, name)
        except AttributeError:
            module = object.__getattribute__(cls, "MODULE")
            for attempt in [
                partial(pointer_class, module=module),
                lambda name: module.__dict__.get(name, AttributeError),
            ]:
                prop = attempt(name)
                if prop is not AttributeError:
                    return prop
            raise

    @classmethod
    def register_call_mod(cls, mod):
        cls.CALL_MODS.add(mod)
        return mod

    @classmethod
    def wrap(cls, func):
        sig = inspect.signature(func)
        parameters = list(sig.parameters.values())

        @wraps(func)
        def wrapper(*args, **kwargs):
            """
            wrapper will be assigned to the ESYSContext class as a method. As
            such the first argument, self, is an instance of ESYSContext
            """
            args = list(args)
            # Combine the arguments we were passed and the parameters from the
            # signature and loop through them all.
            for i, (value, parameter) in enumerate(zip(args, parameters)):
                # Go through each of the call modifiers and use the returned
                # value as the new value for the argument if it was not None
                for modify in cls.CALL_MODS:
                    modifed = modify(parameter.name, parameter.annotation, value)
                    if modifed is not None:
                        args[i] = modifed
            LOGGER.debug(
                ("%s(" % (func.__name__,))
                + ("\n    " if args else "")
                + "\n    ".join(
                    map(lambda x: "%s: %s," % (x[0].name, x[1]), zip(parameters, args))
                )
                + ("\n" if args else "")
                + ")"
            )
            return func(*args, **kwargs)

        return wrapper

    @staticmethod
    def call_mod_ptr_or_value(annotation, value):
        """
        Last step in a call_mod_ for classes which wrap swig types and expose them
        via ``value`` and ``ptr`` properties.
        """
        # If a pointer is being requested, then pass the SessionContext pointer. Do
        # this by checking if the reverse of the string representation of the value
        # starts in a *, aka the last charater in the type is a * (for pointer)
        if annotation[::-1].startswith("*"):
            return value.ptr
        # Otherwise we pass the value that is being pointed to by the SessionContext
        # pointer
        return value.value


@WrapperMetaClass.register_call_mod
def call_mod_context_managed_pointer_class(name, annotation, value):
    if isinstance(value, ContextManagedPointerClass):
        return WrapperMetaClass.call_mod_ptr_or_value(annotation, value)



import ctypes
from contextlib import ExitStack

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa

from tpm2_pytss.binding import *
from tpm2_pytss.util.testing import BaseTestESYS


class TestLoadExternal(BaseTestESYS):
    def test_load_public_rsa_key(self):
        # The tpm2-tools equivalent of what this tests is attempting to do is
        # the following:

        # tpm2_loadexternal
        #    --hierarchy=o
        #    --key-algorithm=rsa
        #    --public=public_rsa_key.pem
        #    --key-context=signing_key.ctx
        #    --name=signing_key.name

        # Create the 2048 bit RSA key.
        private_rsa_key = rsa.generate_private_key(
            public_exponent=65537, key_size=2048, backend=default_backend()
        )
        public_rsa_key = private_rsa_key.public_key()

        with ExitStack() as stack:
            # Create the public part.
            in_public = TPM2B_PUBLIC(
                size=0,
                publicArea=TPMT_PUBLIC(
                    type=TPM2_ALG_RSA,  # key-algorithm=rsa
                    nameAlg=TPM2_ALG_SHA256,  # hash-algorithm = default = sha256
                    objectAttributes=ctypes.c_uint32(
                        TPMA_OBJECT_NODA
                        | TPMA_OBJECT_DECRYPT
                        | TPMA_OBJECT_SIGN_ENCRYPT
                        | TPMA_OBJECT_USERWITHAUTH
                    ).value,
                    authPolicy=TPM2B_DIGEST(size=0),
                    parameters=TPMU_PUBLIC_PARMS(
                        rsaDetail=TPMS_RSA_PARMS(
                            scheme=TPMT_RSA_SCHEME(
                                scheme=TPM2_ALG_NULL,
                                details=TPMU_ASYM_SCHEME(
                                    anySig=TPMS_SCHEME_HASH(hashAlg=TPM2_ALG_NULL,),
                                ),
                            ),
                            symmetric=TPMT_SYM_DEF_OBJECT(
                                algorithm=TPM2_ALG_NULL,
                                keyBits=TPMU_SYM_KEY_BITS(sym=0),
                                mode=TPMU_SYM_MODE(sym=TPM2_ALG_NULL),
                            ),
                            keyBits=public_rsa_key.key_size,  # Number of bits in the key 2048
                            exponent=public_rsa_key.public_numbers().e,  # exponent of the key 65537
                        ),
                    ),
                    unique=TPMU_PUBLIC_ID(
                        rsa=TPM2B_PUBLIC_KEY_RSA(
                            # public part of rsa key.
                            buffer=public_rsa_key.public_numbers().n.to_bytes(
                                public_rsa_key.key_size // 8, byteorder="big"
                            )
                        )
                    ),
                ),
            )

            in_public_ptr = stack.enter_context(in_public.ptr())
            object_handle = stack.enter_context(self.esys_ctx.flush_tr())

            self.esys_ctx.LoadExternal(
                ESYS_TR_NONE,
                ESYS_TR_NONE,
                ESYS_TR_NONE,
                None,
                in_public_ptr,
                TPM2_RH_OWNER,
                object_handle,
            )

            # Get name from the object_handle and compare it to the public key.
            # If there is a difference throw an exception.



import os
import time
import tempfile
import unittest
import contextlib

from .. import tcti
from ..esys import ESYS
from ..fapi import FAPI, FAPIDefaultConfig
from ..exceptions import TPM2Error
from .simulator import SimulatorTest
from .retry import TCTI_RETRY_TRIES

ENV_TCTI = "PYESYS_TCTI"
ENV_TCTI_DEFAULT = "mssim"
ENV_TCTI_CONFIG = "PYESYS_TCTI_CONFIG"
ENV_TCTI_CONFIG_DEFAULT = None


class BaseTestESYS(SimulatorTest, unittest.TestCase):
    """
    ESYS tests should subclass from this
    """

    def setUp(self):
        super().setUp()
        self.esys = ESYS()
        self.tcti = tcti.TCTI.load(os.getenv(ENV_TCTI, default=ENV_TCTI_DEFAULT))
        self.tcti_config = os.getenv(
            ENV_TCTI_CONFIG, default="port=%d" % (self.simulator.port)
        )
        # Create a context stack
        self.ctx_stack = contextlib.ExitStack().__enter__()
        # Enter the contexts
        self.tcti_ctx = self.ctx_stack.enter_context(
            self.tcti(config=self.tcti_config, retry=TCTI_RETRY_TRIES)
        )
        self.esys_ctx = self.ctx_stack.enter_context(self.esys(self.tcti_ctx))
        # Call Startup and clear the TPM
        self.esys_ctx.Startup(self.esys_ctx.TPM2_SU_CLEAR)
        # Set the timeout to blocking
        self.esys_ctx.SetTimeout(self.esys_ctx.TSS2_TCTI_TIMEOUT_BLOCK)

    def tearDown(self):
        super().tearDown()
        self.ctx_stack.__exit__(None, None, None)


class BaseTestFAPI(SimulatorTest, unittest.TestCase):
    """
    FAPI tests should subclass from this
    """

    def setUp(self):
        super().setUp()
        # Create a context stack
        self.ctx_stack = contextlib.ExitStack().__enter__()
        # Create temporary directories
        self.user_dir = self.ctx_stack.enter_context(tempfile.TemporaryDirectory())
        self.log_dir = self.ctx_stack.enter_context(tempfile.TemporaryDirectory())
        self.system_dir = self.ctx_stack.enter_context(tempfile.TemporaryDirectory())
        # Create the FAPI object
        self.fapi = FAPI(
            FAPIDefaultConfig._replace(
                user_dir=self.user_dir,
                system_dir=self.system_dir,
                log_dir=self.log_dir,
                tcti="mssim:port=%d" % (self.simulator.port,),
                tcti_retry=TCTI_RETRY_TRIES,
                ek_cert_less=1,
            )
        )
        # Enter the context
        self.fapi_ctx = self.ctx_stack.enter_context(self.fapi)

    def tearDown(self):
        super().tearDown()
        self.ctx_stack.__exit__(None, None, None)



import os
import ast
import sys
import pathlib
import unittest
import subprocess


# Top of the git repo
ROOT = pathlib.Path(__file__).parent.parent

ENV = os.environ.copy()
ENV["PYTHONPATH"] = str(ROOT)


def filepath(filename):
    return ROOT / "examples" / filename


class TestExamples(unittest.TestCase):
    def test_fapi_get_random(self):
        # Capture output
        stdout = (
            subprocess.check_output(
                [sys.executable, str(filepath("fapi_get_random.py"))], env=ENV,
            )
            .decode()
            .strip()
        )
        # Parse the length
        length = int(ast.literal_eval(stdout.split(")")[0].split("(")[-1]))
        # Parse the value of the bytearray
        array = bytearray(ast.literal_eval(stdout.split("(")[-1].replace(")", "")))
        # Ensure the bytearray makes sense
        self.assertEqual(length, len(array))



import time
import unittest
import unittest.mock

from tpm2_pytss import tcti
from tpm2_pytss.exceptions import TPM2Error
from tpm2_pytss.util.retry import retry_tcti_loop, retry_tcti_catch, TCTI_RETRY_TRIES


class TestTCTIRetry(unittest.TestCase):
    def test_tcti_retry_failure(self):
        self.tcti = tcti.TCTI.load("mssim")
        self.tcti_config = "port=-1"
        with unittest.mock.patch(
            "tpm2_pytss.esys.ESYSBinding.Tss2_TctiLdr_Initialize_Ex",
            side_effect=TPM2Error(655370),
        ), unittest.mock.patch("time.sleep", return_value=True):
            with self.assertRaises(TPM2Error):
                for retry in retry_tcti_loop():
                    with retry_tcti_catch(retry):
                        with self.tcti(config=self.tcti_config) as _tcti_ctx:
                            pass
            self.assertEqual(retry.i, TCTI_RETRY_TRIES)

    def test_tcti_retry_success(self):
        self.tcti = tcti.TCTI.load("mssim")
        self.tcti_config = "port=-1"
        with unittest.mock.patch(
            "tpm2_pytss.esys.ESYSBinding.Tss2_TctiLdr_Initialize_Ex", return_value=0
        ):
            for retry in retry_tcti_loop():
                with retry_tcti_catch(retry):
                    with self.tcti(config=self.tcti_config) as _tcti_ctx:
                        pass
            self.assertEqual(retry.i, 0)



# SPDX-License-Identifier: MIT
# Copyright (c) 2019 Intel Corporation
# All rights reserved.
import os
import atexit
import random
import asyncio
import subprocess


class SimulatorFailedToStart(Exception):
    pass  # pragma: no cov


class Simulator:

    HOST = "127.0.0.1"

    def __init__(self, binary="tpm_server", host=None, port=None):
        self.binary = binary
        self.host = host if host is not None else self.HOST
        self.port = port if port is not None else random.randrange(49152, 65535)
        self.proc = None

    async def _create_connection_after(self, loop, delay):
        await asyncio.sleep(delay, loop=loop)
        try:
            reader, writer = await asyncio.open_connection(
                self.host, self.port, loop=loop
            )
            writer.close()
            return True
        except Exception as error:
            # Don't raise connection errors, return false if we didn't connect
            return False

    async def _stagger_create_connections(self, loop, timeout, stagger=0.05):
        complete = False
        # Stagger the creation of connections
        tasks = set(
            [
                loop.create_task(self._create_connection_after(loop, delay * stagger))
                for delay in range(0, int(timeout / stagger))
            ]
        )
        try:
            while tasks:
                done, pending = await asyncio.wait(
                    tasks, return_when=asyncio.FIRST_COMPLETED, loop=loop
                )
                # Check if the done coroutines that finished raised exceptions
                for task in done:
                    failed = task.exception()
                    if failed is not None:
                        raise failed
                complete = any([task.result() for task in done])
                if complete:
                    break
        finally:
            # Cancel all unconnected coroutines
            for task in tasks:
                if not task.done():
                    task.cancel()
        return complete

    async def _connect_to_simulator(self, loop, timeout):
        done, pending = await asyncio.wait(
            [loop.create_task(self._stagger_create_connections(loop, timeout))],
            timeout=timeout,
            loop=loop,
        )
        # If none are done then we got a timeout
        if not done:
            raise TimeoutError(
                "Could not connect to simulator within {} seconds".format(timeout)
            )
        # Cancel all unconnected coroutines
        for task in pending:
            task.cancel()
        # Check if the done coroutines that finished raised exceptions
        for task in done:
            failed = task.exception()
            if failed is not None:
                raise failed

    def connect_to_simulator(self, timeout):
        loop = asyncio.new_event_loop()
        loop.run_until_complete(self._connect_to_simulator(loop, timeout))
        loop.close()

    def start(self, timeout=2):
        # Remove simulator state if it exists
        if os.path.isfile("NVChip"):
            os.unlink("NVChip")
        # We are setting up a simulator in here, because we require the standard
        # ports. pytpm2tss does not yet support setting tctis and thus relies on
        # esys to do so.
        self.proc = subprocess.Popen(
            [self.binary, "-rm", "-port", str(self.port)],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        atexit.register(self.stop)
        try:
            self.connect_to_simulator(timeout)
        except TimeoutError as error:
            raise SimulatorFailedToStart(self.output) from error
        if self.proc.poll() is not None:
            raise SimulatorFailedToStart(self.output)

    @property
    def output(self):
        if self.proc is not None:
            stderr, stdout = self.proc.communicate()
            return "{}\n{}".format(stderr.decode(), stdout.decode()).strip()
        return "Not running"

    def stop(self):
        atexit.unregister(self.stop)
        if self.proc is not None:
            self.proc.stdout.close()
            self.proc.stderr.close()
            self.proc.terminate()
            self.proc.wait()
            self.proc = None

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, _exc_type, _exc_value, _traceback):
        self.stop()


class SimulatorTest:
    """
    Testcase that requires the tpm_server simulator to be running.
    """

    def setUp(self):
        super().setUp()
        self.simulator = Simulator()
        if not os.getenv("SIM_RUNNING", default=""):
            self.simulator.start()

    def tearDown(self):
        super().tearDown()
        if not os.getenv("SIM_RUNNING", default=""):
            self.simulator.stop()



import random
from contextlib import ExitStack

from tpm2_pytss.binding import *
from tpm2_pytss.esys import InvalidArgumentError
from tpm2_pytss.util.testing import BaseTestESYS


class TestGetRandom(BaseTestESYS):
    def test_random_length(self):
        length = random.randint(8, 32)

        array = self.esys_ctx.get_random(length)

        self.assertEqual(length, len(array))

    def test_invalid_length(self):
        with self.assertRaises(InvalidArgumentError):
            self.esys_ctx.get_random(65)

    def test_start_auth_session(self):
        with ExitStack() as stack:

            symmetric = TPMT_SYM_DEF(
                algorithm=TPM2_ALG_AES,
                keyBits=TPMU_SYM_KEY_BITS(aes=128),
                mode=TPMU_SYM_MODE(aes=TPM2_ALG_CFB),
            )

            symmetric_ptr = stack.enter_context(symmetric.ptr())

            session = stack.enter_context(
                self.esys_ctx.auth_session(
                    ESYS_TR_NONE,
                    ESYS_TR_NONE,
                    ESYS_TR_NONE,
                    ESYS_TR_NONE,
                    ESYS_TR_NONE,
                    None,
                    TPM2_SE_HMAC,
                    symmetric_ptr,
                    TPM2_ALG_SHA1,
                )
            )

            self.esys_ctx.TRSess_SetAttributes(
                session, TPMA_SESSION_AUDIT, TPMA_SESSION_AUDIT
            )

            length = 48

            array = self.esys_ctx.get_random(length, shandle1=session)

            self.assertEqual(length, len(array))



import os
import time
import tempfile
import unittest
import contextlib

from tpm2_pytss import tcti
from tpm2_pytss.esys import ESYS
from tpm2_pytss.fapi import FAPI, FAPIDefaultConfig
from tpm2_pytss.exceptions import TPM2Error
from tpm2_pytss.util.simulator import SimulatorTest
from tpm2_pytss.util.retry import TCTI_RETRY_TRIES

ENV_TCTI = "PYESYS_TCTI"
ENV_TCTI_DEFAULT = "mssim"
ENV_TCTI_CONFIG = "PYESYS_TCTI_CONFIG"
ENV_TCTI_CONFIG_DEFAULT = None


class BaseTestESYS(SimulatorTest, unittest.TestCase):
    """
    ESYS tests should subclass from this
    """

    def setUp(self):
        super().setUp()
        self.esys = ESYS()
        self.tcti = tcti.TCTI.load(os.getenv(ENV_TCTI, default=ENV_TCTI_DEFAULT))
        self.tcti_config = os.getenv(
            ENV_TCTI_CONFIG, default="port=%d" % (self.simulator.port)
        )
        # Create a context stack
        self.ctx_stack = contextlib.ExitStack().__enter__()
        # Enter the contexts
        self.tcti_ctx = self.ctx_stack.enter_context(
            self.tcti(config=self.tcti_config, retry=TCTI_RETRY_TRIES)
        )
        self.esys_ctx = self.ctx_stack.enter_context(self.esys(self.tcti_ctx))
        # Call Startup and clear the TPM
        self.esys_ctx.Startup(self.esys_ctx.TPM2_SU_CLEAR)
        # Set the timeout to blocking
        self.esys_ctx.SetTimeout(self.esys_ctx.TSS2_TCTI_TIMEOUT_BLOCK)

    def tearDown(self):
        super().tearDown()
        self.ctx_stack.__exit__(None, None, None)


class BaseTestFAPI(SimulatorTest, unittest.TestCase):
    """
    FAPI tests should subclass from this
    """

    def setUp(self):
        super().setUp()
        # Create a context stack
        self.ctx_stack = contextlib.ExitStack().__enter__()
        # Create temporary directories
        self.user_dir = self.ctx_stack.enter_context(tempfile.TemporaryDirectory())
        self.log_dir = self.ctx_stack.enter_context(tempfile.TemporaryDirectory())
        self.system_dir = self.ctx_stack.enter_context(tempfile.TemporaryDirectory())
        # Create the FAPI object
        self.fapi = FAPI(
            FAPIDefaultConfig._replace(
                user_dir=self.user_dir,
                system_dir=self.system_dir,
                log_dir=self.log_dir,
                tcti="mssim:port=%d" % (self.simulator.port,),
                tcti_retry=TCTI_RETRY_TRIES,
            )
        )
        # Enter the context
        self.fapi_ctx = self.ctx_stack.enter_context(self.fapi)

    def tearDown(self):
        super().tearDown()
        self.ctx_stack.__exit__(None, None, None)



import os
import ast
import stat
import glob
import json
import shlex
import shutil
import pathlib
from io import open
from subprocess import check_output
from setuptools.command.build_py import build_py
from setuptools.command.build_ext import build_ext
from setuptools import find_packages, setup, Extension

ORG = "tpm2-software"
NAME = "tpm2-pytss"
DESCRIPTION = "TPM 2.0 TSS Bindings for Python"
AUTHOR_NAME = "John Andersen"
AUTHOR_EMAIL = "john.s.andersen@intel.com"
INSTALL_REQUIRES = []

IMPORT_NAME = NAME.replace("-", "_")

SELF_PATH = os.path.dirname(os.path.realpath(__file__))

with open(os.path.join(SELF_PATH, IMPORT_NAME, "version"), "r") as f:
    VERSION = "".join(f).strip()

README = pathlib.Path(SELF_PATH, "README.md").read_text()


class PkgConfigNeededExtension(Extension):
    """
    By creating a subclass of Extension and using the :py:func:property builtin
    we can make it so that pkg-config doesn't get called unless the user
    attempts to build from source. Otherwise we'd run into issues installing the
    built binary on systems that don't have pkg-config.
    """

    def __init__(self, *args, pkg_config_cflags=None, pkg_config_libs=None, **kwargs):
        # Default to empty array if not given
        if pkg_config_cflags is None:
            pkg_config_cflags = []
        if pkg_config_libs is None:
            pkg_config_libs = []
        self.pkg_config_cflags = pkg_config_cflags
        self.pkg_config_libs = pkg_config_libs
        # Will be populated by respective non-underscore setters
        self._libraries = []
        self._include_dirs = []
        self._library_dirs = []
        self._swig_opts = []
        super().__init__(*args, **kwargs)

    @property
    def cc_include_dirs(self):
        if not self.pkg_config_cflags:
            return []
        return shlex.split(
            check_output(
                ["pkg-config", "--cflags"] + self.pkg_config_cflags,
                env=dict(os.environ, PKG_CONFIG_ALLOW_SYSTEM_CFLAGS="1"),
            ).decode()
        )

    @property
    def cc_libraries(self):
        if not self.pkg_config_libs:
            return []
        return shlex.split(
            check_output(["pkg-config", "--libs"] + self.pkg_config_libs).decode()
        )

    def _strip_leading(self, number, iterable):
        """
        Strips number characters from the begining of each string in iterable.
        """
        return list(map(lambda string: string[number:], iterable))

    def _subset_startswith(self, prefix, iterable):
        """
        Subset of list where each element starts with prefix.
        """
        return list(filter(lambda option: option.startswith(prefix), iterable))

    def _strip_only_startswith(self, prefix, iterable):
        """
        Subset of list where each element starts with prefix, where prefix has
        now been removed.
        """
        return self._strip_leading(
            len(prefix), self._subset_startswith(prefix, iterable)
        )

    @property
    def include_dirs(self):
        return (
            self._strip_only_startswith("-I", self.cc_include_dirs) + self._include_dirs
        )

    @include_dirs.setter
    def include_dirs(self, value):
        self._include_dirs = value

    @property
    def libraries(self):
        return self._strip_only_startswith("-l", self.cc_libraries) + self._libraries

    @libraries.setter
    def libraries(self, value):
        self._libraries = value

    @property
    def library_dirs(self):
        return self._strip_only_startswith("-L", self.cc_libraries) + self._library_dirs

    @library_dirs.setter
    def library_dirs(self, value):
        self._library_dirs = value

    @property
    def swig_opts(self):
        return self._subset_startswith("-I", self.cc_include_dirs) + self._swig_opts

    @swig_opts.setter
    def swig_opts(self, value):
        self._swig_opts = value


class BuildSWIGPyFiles(build_py):
    def run(self):
        # This is a workaround for https://bugs.python.org/issue37247
        # Run regular copy of files, required to run SWIG build
        super().run()
        # Causes SWIG python files to be built. Results in SWIG build happening
        # twice, maybe but not sure
        cmd = BuildExtThenCopySWIGPy(self.distribution)
        cmd.finalize_options()
        cmd.run()
        # Run copy of py files again, now that SWIG has generated py files
        super().run()


class BuildExtThenCopySWIGPy(build_ext):
    def run(self):
        super().run()
        # SWIG 4 Support
        for fixfile in ["esys_binding.py", "fapi_binding.py"]:
            binding_path = pathlib.Path(SELF_PATH, IMPORT_NAME, fixfile)
            binding = binding_path.read_text()
            if not "python_property = property" in binding:
                binding = binding.replace("= property(", "= python_property(")
                binding = "python_property = property\n" + binding
                binding_path.write_text(binding)
        # This is needed because test copies the binding files into IMPORT_NAME
        # but build does not. Making this necessary for working with the package
        # installed in development mode.
        for src in glob.glob(
            os.path.join(SELF_PATH, "build", "lib.*", "**", "*.so"), recursive=True
        ):
            dst = os.path.join(SELF_PATH, IMPORT_NAME, os.path.basename(src))
            print("{} -> {}".format(src, dst))
            shutil.copyfile(src, dst)
            os.chmod(
                dst,
                stat.S_IRUSR
                | stat.S_IWUSR
                | stat.S_IXUSR
                | stat.S_IRGRP
                | stat.S_IXGRP
                | stat.S_IROTH
                | stat.S_IXOTH,
            )


setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=README,
    long_description_content_type="text/markdown",
    author=AUTHOR_NAME,
    author_email=AUTHOR_EMAIL,
    maintainer=AUTHOR_NAME,
    maintainer_email=AUTHOR_EMAIL,
    url="https://github.com/{}/{}".format(ORG, NAME),
    license="MIT",
    keywords=["tpm2", "security"],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    install_requires=INSTALL_REQUIRES,
    tests_require=["cryptography>=2.8"],
    extras_require={
        "dev": [
            "coverage",
            "codecov",
            "sphinx",
            "sphinxcontrib-asyncio",
            "black",
            "sphinx_rtd_theme",
        ],
    },
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    ext_modules=[
        PkgConfigNeededExtension(
            "{}._esys_binding".format(IMPORT_NAME),
            [os.path.join(IMPORT_NAME, "swig", "esys_binding.i")],
            pkg_config_cflags=["tss2-esys", "tss2-rc", "tss2-tctildr"],
            pkg_config_libs=["tss2-esys", "tss2-rc", "tss2-tctildr"],
            swig_opts=["-py3", "-outdir", IMPORT_NAME],
        ),
        PkgConfigNeededExtension(
            "{}._fapi_binding".format(IMPORT_NAME),
            [os.path.join(IMPORT_NAME, "swig", "fapi_binding.i")],
            pkg_config_cflags=["tss2-fapi", "tss2-rc", "tss2-tctildr"],
            pkg_config_libs=["tss2-fapi", "tss2-rc", "tss2-tctildr"],
            swig_opts=["-py3", "-outdir", IMPORT_NAME],
        ),
    ],
    py_modules=[IMPORT_NAME],
    cmdclass={"build_ext": BuildExtThenCopySWIGPy, "build_py": BuildSWIGPyFiles},
    include_package_data=True,
    zip_safe=False,
)



import random
from contextlib import ExitStack

from tpm2_pytss.binding import *
from tpm2_pytss.util.testing import BaseTestFAPI


class TestGetRandom(BaseTestFAPI):
    def test_get_info(self):
        self.fapi_ctx.Provision(None, None, None)

        with CHAR_PTR_PTR() as array:
            self.assertEqual(type(self.fapi_ctx.GetInfo(array)), dict)

    def test_random_length(self):
        length = random.randint(8, 32)

        self.fapi_ctx.Provision(None, None, None)

        with UINT8_PTR_PTR() as array:
            array = self.fapi_ctx.GetRandom(length, array)

            value = to_bytearray(length, array)

            self.assertEqual(length, len(value))



# SPDX-License-Identifier: MIT
# Copyright (c) 2019 Intel Corporation
from functools import partial
from typing import Optional, ByteString

from .tcti import TCTIContext
from .util.swig import Wrapper
from .binding import AuthSessionContext, ESYSBinding, FlushTRContext, NVContext
from .context import BaseContextMetaClass


class InvalidArgumentError(Exception):
    pass  # pragma: no cov


class ESYSContextMetaClass(BaseContextMetaClass):
    PREFIX = "Esys_"
    NO_PASS_CTXP = set(["Esys_Initialize", "Esys_Finalize"])


class ESYSContext(Wrapper, metaclass=ESYSContextMetaClass):

    MODULE = ESYSBinding
    # TODO Verify maximum length requested of GetRandom
    GET_RANDOM_MAX_LENGTH = 64

    def __init__(self, parent: "ESYS", tcti_ctx: TCTIContext) -> None:
        self.parent = parent
        self.tcti_ctx = tcti_ctx
        self.ctxpp = None
        self.ctxp = None
        # self.logger.debug("__init__")

    def __enter__(self) -> "ESYSContext":
        # Create an ESYS_CONTEXT **
        ctxpp = self.new_ctx_ptr()
        self.Initialize(ctxpp, self.tcti_ctx.ctxp, self.parent.abi_version)
        # Grab the allocated ESYS_CONTEXT *
        ctxp = self.ctx_ptr_value(ctxpp)
        # Save references at the end to avoid possible memory leaks
        self.ctxpp = ctxpp
        self.ctxp = ctxp
        # self.logger.debug("__enter__")
        return self

    def __exit__(self, _exc_type, _exc_value, _traceback):
        self.Finalize(self.ctxpp)
        self.delete_ctx_ptr(self.ctxpp)
        self.ctxpp = None
        self.ctxp = None
        # self.logger.debug("__exit__")

    def _bytearray(self, length, buf):
        """
        Takes any pointer to an array of bytes and that array's length and
        returns a :py:`bytearray` containing those bytes.
        """
        buf = ESYSBinding.ByteArray.frompointer(buf)
        array = bytearray(length)
        for i in range(0, length):
            array[i] = buf[i]
        return array

    @property
    def flush_tr(self):
        """
        Create and return a ESYS_TR_PTR that will be set to ESYS_TR_NONE and
        flushed when it's context exits.
        """
        return partial(FlushTRContext, self)

    @property
    def auth_session(self):
        """
        Create and return a auth session context.
        """
        return partial(AuthSessionContext, self)

    @property
    def nv(self):
        """
        Create and return a nv context.
        """
        return partial(NVContext, self)

    def get_random(
        self,
        length: int,
        shandle1=ESYSBinding.ESYS_TR_NONE,
        shandle2=ESYSBinding.ESYS_TR_NONE,
        shandle3=ESYSBinding.ESYS_TR_NONE,
    ) -> ByteString:
        """
        Esys_GetRandom
        """
        if length > self.GET_RANDOM_MAX_LENGTH:
            raise InvalidArgumentError(
                "Maximum length is {}".format(self.GET_RANDOM_MAX_LENGTH)
            )

        with self.TPM2B_DIGEST_PTR_PTR() as datapp:
            self.GetRandom(shandle1, shandle2, shandle3, length, datapp)

            return self._bytearray(datapp.value.size, datapp.value.buffer)


class ESYS(Wrapper):

    MODULE = ESYSBinding
    CONTEXT = ESYSContext
    TSS_CREATOR = 1
    TSS_FAMILY = 2
    TSS_LEVEL = 1
    TSS_VERSION = 108

    def __init__(
        self, *, abi_version: Optional[ESYSBinding.TSS2_ABI_VERSION] = None
    ) -> None:
        self.abi_version = (
            abi_version if abi_version is not None else self.default_abi_version()
        )
        # self.logger.debug("__init__")

    def __call__(self, tcti_ctx: TCTIContext) -> "ESYSContext":
        return self.CONTEXT(self, tcti_ctx)

    @classmethod
    def default_abi_version(cls):
        version = cls.MODULE.TSS2_ABI_VERSION()
        version.tssCreator = cls.TSS_CREATOR
        version.tssFamily = cls.TSS_FAMILY
        version.tssLevel = cls.TSS_LEVEL
        version.tssVersion = cls.TSS_VERSION
        # self.logger.debug("created instance of default abi version: %r", version)
        return version



# SPDX-License-Identifier: MIT
# Copyright (c) 2019 Intel Corporation
import json
import inspect
import contextlib
from functools import partial, wraps
from typing import Optional, ByteString

from .tcti import TCTIContext
from .util.swig import Wrapper
from .binding import (
    AuthSessionContext,
    ESYSBinding,
    FlushTRContext,
    NVContext,
    to_bytearray,
)


class InvalidArgumentError(Exception):
    pass  # pragma: no cov


class BaseContextMetaClass(type):
    def __new__(cls, name, bases, props):
        # Create the class
        cls = super(BaseContextMetaClass, cls).__new__(cls, name, bases, props)
        # Go through all the functions in the module
        module = props["MODULE"].__dict__
        for key, func in module.items():
            if key.startswith("_"):
                continue
            # Check if a custom wrapper has been defined for this function
            custom_wrap = getattr(cls, "wrap_{}".format(key), None)
            if custom_wrap is not None:
                func = custom_wrap(func)
            # For all Esys_ functions wrap the function and make the wrapped
            # function a method in our new class
            elif key.startswith(cls.PREFIX):
                if key in cls.NO_PASS_CTXP:
                    func = cls._wrap_no_pass_ctxp(func)
                else:
                    func = cls._wrap_pass_ctxp(func)
            # Remove Esys_ from function names since they will be called from
            # the ESYSContext and that's redundant
            if key.startswith(cls.PREFIX):
                key = key[len(cls.PREFIX) :]
            setattr(cls, key, func)
        return cls

    @classmethod
    def _wrap_no_pass_ctxp(cls, func):
        @wraps(func)
        def wrapper(_self, *args, **kwds):
            """
            wrapper will be assigned to the ESYSContext class as a method. As
            such the first argument, self, is an instance of ESYSContext.
            """
            # Check if a custom wrapper has been defined
            custom_wrap = getattr(cls, "wrap_no_pass_ctxp", None)
            if custom_wrap is not None:
                return custom_wrap(func)(*args, **kwds)
            return func(*args, **kwds)

        return wrapper

    @classmethod
    def _wrap_pass_ctxp(cls, func):
        @wraps(func)
        def wrapper(self, *args, **kwds):
            """
            wrapper will be assigned to the ESYSContext class as a method. As
            such the first argument, self, is an instance of ESYSContext.
            """
            # Check if a custom wrapper has been defined
            custom_wrap = getattr(cls, "wrap_pass_ctxp", None)
            if custom_wrap is not None:
                return custom_wrap(func)(self.ctxp, *args, **kwds)
            return func(self.ctxp, *args, **kwds)

        return wrapper

    @classmethod
    def wrap_pass_ctxp(cls, func):
        """
        Take out pointers and make them tuple return values
        """

        @wraps(func)
        def wrapper(self, *args, **kwds):
            sig = inspect.signature(func.orig)
            parameters = list(sig.parameters.values())

            docstring_arguments = func.orig.__doc__.split("(")[1].split(")")[0]
            # Remove FAPI/ESAPI context pointer
            docstring_arguments = docstring_arguments.split(",")[1:]
            docstring_arguments = map(lambda i: i.strip(), docstring_arguments)
            docstring_arguments = list(docstring_arguments)

            # First missing argument should not be const
            if (
                len(args) < len(docstring_arguments)
                and "const" in docstring_arguments[len(args) :][0].split()
            ):
                for i, docstring in enumerate(docstring_arguments[len(args) :]):
                    # Ensure that all these are arguments that need
                    # allocating
                    if not "**" in docstring:
                        break
                    print(docstring)

            result = func(self, *args, **kwds)

            return_value = []

            skip = False
            for i, (value, docstring) in enumerate(zip(args, docstring_arguments)):
                if skip:
                    skip = False
                    continue
                if not "const" in docstring.split() and "*" in docstring:
                    # char ** arguments are always guaranteed to be NULL
                    # terminated
                    if docstring.startswith("char") and "**" in docstring.split():
                        if any(map(value.value.startswith, ["[", "{"])):
                            try:
                                return_value.append(json.loads(value.value))
                            except json.decoder.JSONDecodeError:
                                return_value.append(value.value)
                    elif "**" in docstring.split():
                        if (i + 1) < len(docstring_arguments):
                            next_docstring = docstring_arguments[i + 1]
                            next_docstring = next_docstring.split()
                            if (
                                "size_t" in next_docstring
                                and "*" in next_docstring
                                and "uint8_t" in docstring.split()
                            ):
                                return_value.append(
                                    to_bytearray(value.value, args[i + 1].value)
                                )
                                skip = True
                                continue
                        return_value.append(value.value)
                    else:
                        return_value.append(value)

            if len(return_value) > 1:
                return return_value
            elif len(return_value) == 1:
                return return_value[0]
            return result

        return wrapper


class ESYSContextMetaClass(BaseContextMetaClass):
    PREFIX = "Esys_"
    NO_PASS_CTXP = set(["Esys_Initialize", "Esys_Finalize"])


class ESYSContext(Wrapper, metaclass=ESYSContextMetaClass):

    MODULE = ESYSBinding
    # TODO Verify maximum length requested of GetRandom
    GET_RANDOM_MAX_LENGTH = 64

    def __init__(self, parent: "ESYS", tcti_ctx: TCTIContext) -> None:
        self.parent = parent
        self.tcti_ctx = tcti_ctx
        self.ctxpp = None
        self.ctxp = None
        # self.logger.debug("__init__")

    def __enter__(self) -> "ESYSContext":
        # Create an ESYS_CONTEXT **
        ctxpp = self.new_ctx_ptr()
        self.Initialize(ctxpp, self.tcti_ctx.ctxp, self.parent.abi_version)
        # Grab the allocated ESYS_CONTEXT *
        ctxp = self.ctx_ptr_value(ctxpp)
        # Save references at the end to avoid possible memory leaks
        self.ctxpp = ctxpp
        self.ctxp = ctxp
        # self.logger.debug("__enter__")
        return self

    def __exit__(self, _exc_type, _exc_value, _traceback):
        self.Finalize(self.ctxpp)
        self.delete_ctx_ptr(self.ctxpp)
        self.ctxpp = None
        self.ctxp = None
        # self.logger.debug("__exit__")

    def _bytearray(self, length, buf):
        """
        Takes any pointer to an array of bytes and that array's length and
        returns a :py:`bytearray` containing those bytes.
        """
        buf = ESYSBinding.ByteArray.frompointer(buf)
        array = bytearray(length)
        for i in range(0, length):
            array[i] = buf[i]
        return array

    @property
    def flush_tr(self):
        """
        Create and return a ESYS_TR_PTR that will be set to ESYS_TR_NONE and
        flushed when it's context exits.
        """
        return partial(FlushTRContext, self)

    @property
    def auth_session(self):
        """
        Create and return a auth session context.
        """
        return partial(AuthSessionContext, self)

    @property
    def nv(self):
        """
        Create and return a nv context.
        """
        return partial(NVContext, self)

    def get_random(
        self,
        length: int,
        shandle1=ESYSBinding.ESYS_TR_NONE,
        shandle2=ESYSBinding.ESYS_TR_NONE,
        shandle3=ESYSBinding.ESYS_TR_NONE,
    ) -> ByteString:
        """
        Esys_GetRandom
        """
        if length > self.GET_RANDOM_MAX_LENGTH:
            raise InvalidArgumentError(
                "Maximum length is {}".format(self.GET_RANDOM_MAX_LENGTH)
            )

        with self.TPM2B_DIGEST_PTR_PTR() as datapp:
            self.GetRandom(shandle1, shandle2, shandle3, length, datapp)

            return self._bytearray(datapp.value.size, datapp.value.buffer)


class ESYS(Wrapper):

    MODULE = ESYSBinding
    CONTEXT = ESYSContext
    TSS_CREATOR = 1
    TSS_FAMILY = 2
    TSS_LEVEL = 1
    TSS_VERSION = 108

    def __init__(
        self, *, abi_version: Optional[ESYSBinding.TSS2_ABI_VERSION] = None
    ) -> None:
        self.abi_version = (
            abi_version if abi_version is not None else self.default_abi_version()
        )
        # self.logger.debug("__init__")

    def __call__(self, tcti_ctx: TCTIContext) -> "ESYSContext":
        return self.CONTEXT(self, tcti_ctx)

    @classmethod
    def default_abi_version(cls):
        version = cls.MODULE.TSS2_ABI_VERSION()
        version.tssCreator = cls.TSS_CREATOR
        version.tssFamily = cls.TSS_FAMILY
        version.tssLevel = cls.TSS_LEVEL
        version.tssVersion = cls.TSS_VERSION
        # self.logger.debug("created instance of default abi version: %r", version)
        return version



import re
import math

NUMBER = re.compile(r"([0-9]+)")
FORMATS = list(
    map(re.compile, [r"(\(\([0-9]+\+[0-9]+\)\/[0-9]+\))", r"(sizeof\(\w+\))"])
)


def calculate(module, string):
    for i, possible in enumerate(FORMATS):
        if not possible.fullmatch(string):
            continue
        if i == 0:
            one, two, three = map(int, NUMBER.findall(string))
            return math.ceil((one + two) / three)
        elif i == 1:
            type_name = string.replace("sizeof(", "").replace(")", "")
            sizeof_type_name = getattr(module, "sizeof_" + type_name)
            return sizeof_type_name
    return int(string)



# SPDX-License-Identifier: MIT
# Copyright (c) 2019 Intel Corporation
import inspect
from functools import wraps
from typing import Optional, Callable, List, Any

from .esys_binding import (
    Tss2_RC_Decode,
    TPM2_RC_FMT1,
    TPM2_RC_P,
    TPM2_RC_S,
    TPM2_RC_N_MASK,
)


class TPM2Error(Exception):
    def __init__(self, rc: int):
        super(TPM2Error, self).__init__(Tss2_RC_Decode(rc))
        self.rc = rc
        self.handle = 0
        self.parameter = 0
        self.session = 0
        self.error = 0
        if self.rc & TPM2_RC_FMT1:
            self._parse_fmt1()
        else:
            self.error = self.rc

    def _parse_fmt1(self):
        self.error = TPM2_RC_FMT1 + (self.rc & 0x3F)

        if self.rc & TPM2_RC_P:
            self.parameter = (self.rc & TPM2_RC_N_MASK) >> 8
        elif self.rc & TPM2_RC_S:
            self.session = ((self.rc - TPM2_RC_S) & TPM2_RC_N_MASK) >> 8
        else:
            self.handle = (self.rc & TPM2_RC_N_MASK) >> 8


def raise_tpm2_error(func):
    """
    Decorator to wrap TPM2 functions so that non-zero return values trigger
    errors.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        rc = func(*args, **kwargs)
        if isinstance(rc, int) and rc != 0:
            raise TPM2Error(rc)
        return rc

    return wrapper


def raise_tpm2_mu_error(func):
    """
    Decorator to wrap TPM2 MU functions so that non-zero return code values trigger
    errors.
    """

    @wraps(func)
    def mu_wrapper(*args, **kwargs):
        rc, offset = func(*args, **kwargs)
        if isinstance(rc, int) and rc != 0:
            raise TPM2Error(rc)
        return rc, offset

    return mu_wrapper


def wrap_funcs(
    src,
    *,
    dst: Optional[List[Any]] = None,
    cond: Optional[Callable[[Callable], bool]] = None,
    wrapperfunc=raise_tpm2_error
):
    """
    Wrap the functions within the given module with a decorator so that if they
    return integer values that are non-zero an error will be thrown.
    """
    if dst is None:
        dst = [src]
    for key, func in inspect.getmembers(src):
        if inspect.isfunction(func) and (cond is None or cond(key, func)):
            wrapped = wrapperfunc(func)
            for module in dst:
                setattr(module, key, wrapped)
    return dst[0]



from tpm2_pytss.binding import *
from tpm2_pytss.util.testing import BaseTestESYS


class TestGetCapability(BaseTestESYS):
    def test_get_capability(self):
        capability = TPM2_CAP_TPM_PROPERTIES
        prop = TPM2_PT_LOCKOUT_COUNTER
        propertyCount = 1

        with TPMI_YES_NO_PTR(
            False
        ) as moreData_ptr, TPMS_CAPABILITY_DATA_PTR_PTR() as capabilityData_ptr_ptr:
            r = self.esys_ctx.GetCapability(
                ESYS_TR_NONE,
                ESYS_TR_NONE,
                ESYS_TR_NONE,
                capability,
                prop,
                propertyCount,
                moreData_ptr,
                capabilityData_ptr_ptr,
            )
